/**
 * Token class
 * @param type
 * @param value
 * @param position
 * @constructor
 */
var Token = function (type, value, position) {
	this.type = type;
	this.value = value;
	this.position = position;
};

Token.prototype = {
	toString: function () {
		return this.value;
	}
};


/**
 * Lexer class
 * @constructor
 */
var Lexer = function () {
	this.pos = 0;
	this.buffer = '';
	this.tokens = [];
	this.rules = [];
	this.currentToken = 0;
	this.initRules();
};

Lexer.EOF = 1;
Lexer.EQUALS = 2;
Lexer.DEF_PARAM = 3;
Lexer.IDENTIFIER = 4;
Lexer.KW_BEGIN = 5;
Lexer.KW_BEGINGROUP = 6;
Lexer.KW_DEF = 7;
Lexer.KW_END = 8;
Lexer.KW_ENDGROUP = 9;
Lexer.KW_FONTSPEC = 18;
Lexer.KW_SHAPE = 10;
Lexer.KW_SKIP = 11;
Lexer.KW_SPACE = 12;
Lexer.NUMBER = 13;
Lexer.TEXT = 14;
Lexer.UNIT = 15;
Lexer.LEFT_CBRACKET = 16;
Lexer.RIGHT_CBRACKET = 17;
Lexer.tokenNames = [
	'n/a', '<EOF>', '=', 'DEF_PARAM', 'IDENTIFIER', 'BEGIN', 'BEGINGROUP', 'DEF', 'END', 'ENDGROUP',
	'SHAPE', 'SKIP', 'SPACE', 'NUMBER', 'TEXT', 'UNIT', '{', '}', 'FONTSPEC'
];
Lexer.trTokenNames = [
	'n/a', '<EOF>', '=', 'def parametr', 'identifikátor', 'begin', 'begingroup', 'def', 'end', 'endgroup',
	'řez písma (textsc, texttt, textit, textbf nebo textsl)', 'leftskip nebo rightskip', 'vspace', 'číslo', 'text', 'jednotka', '{', '}', 'fontspec'
];

Lexer.prototype = {
	initRules: function () {
		var rules = [
			{type: Lexer.EQUALS, pattern: '='},
			{type: Lexer.DEF_PARAM, pattern: '#[0-9]'},
			{type: Lexer.KW_BEGINGROUP, pattern: '\\\\begingroup'},
			{type: Lexer.KW_BEGIN, pattern: '\\\\begin'},
			{type: Lexer.KW_DEF, pattern: '\\\\def'},
			{type: Lexer.KW_ENDGROUP, pattern: '\\\\endgroup'},
			{type: Lexer.KW_END, pattern: '\\\\end'},
			{type: Lexer.KW_FONTSPEC, pattern: '\\\\fontspec'},
			{type: Lexer.KW_SHAPE, pattern: '\\\\(textsc|texttt|textit|textbf|textsl)'},
			{type: Lexer.KW_SKIP, pattern: '\\\\(leftskip|rightskip)'},
			{type: Lexer.KW_SPACE, pattern: '\\\\vspace'},
			{type: Lexer.NUMBER, pattern: '-?\\d+(\\.\\d+)?'},
			{type: Lexer.UNIT, pattern: '(pt|pc|in|bp|cm|mm|dd|cc|sp|ex|em)'},
			{type: Lexer.LEFT_CBRACKET, pattern: '{'},
			{type: Lexer.RIGHT_CBRACKET, pattern: '}'},
			{type: Lexer.IDENTIFIER, pattern: '\\\\[a-zA-Z]+'},
			{type: Lexer.TEXT, pattern: '[a-zA-Z ]+'}
		];

		for (var i=0; i<rules.length; i++) {
			this.rules.push({
				type: rules[i].type,
				pattern: new RegExp('^' + rules[i].pattern)
			});
		}
	},

	input: function (str) {
		this.buffer = str.replace(/%(.*)$/gm, ''); // odstranit komentare
		this.pos = 0;
	},

	getToken: function () {
		if (this.pos >= this.buffer.length) {
			return null;
		}

		// odstranit whitespace
		var whitespace = new RegExp('\\S', 'g');
		whitespace.lastIndex = this.pos;
		var match = whitespace.exec(this.buffer);
		if (match) {
			this.pos = match.index;
		} else {
			return null;
		}

		for (var i=0; i<this.rules.length; i++) {
			var rule = this.rules[i];
			match = rule.pattern.exec(this.buffer.substr(this.pos));
			if (match) {
				var token = new Token(rule.type, match[0], this.pos);
				this.pos += match[0].length;
				return token;
			}
		}

		// no rule matched
		throw new Error('Neočekávaný znak "' + this.buffer[this.pos] + '" na pozici ' + this.pos);
	},

	tokenize: function () {
		while (true) {
			var token = this.getToken();
			if (!token) {
				break;
			}
			this.tokens.push(token);
		}
		return this.tokens;
	},

	nextToken: function () {
		if (!this.tokens.length) {
			this.tokenize();
		}

		if (this.currentToken >= this.tokens.length) {
			return new Token(Lexer.EOF, '<EOF>');
		}

		var token = this.tokens[this.currentToken];
		this.currentToken++;
		return token;
	}
};


/**
 * Parser class
 * @constructor
 */
var Parser = function () {
	this.lexer = new Lexer();
	this.lookahead = null; // current token
	this.root = null;
	this.currentNode = null;
	this.defParamFound = false;
	this.currentStyle = null;
	this.styles = [];
	this.defaultStyle = {
		name: '',
		font: {
			caps: false,
			face: 'default',
			size: 'normalsize',
			style: 'default',
			type: false
		},
		par: {
			align: 'default',
			hspaceLeft: '0em',
			hspaceRight: '0em',
			vspaceAfter: '0em',
			vspaceBefore: '0em'
		}
	};
};

Parser.prototype = {
	expect: function (x) {
		if (this.lookahead.type === x) {
			this.currentNode.addChild(this.lookahead, this.currentNode);
			this.consume();
		} else {
			throw new Error('Očekáváno "' + Lexer.trTokenNames[x] + '", nalezeno "' + this.lookahead + '"');
		}
	},

	consume: function () {
		this.lookahead = this.lexer.nextToken();
	},

	isLookahead: function (x) {
		if (x instanceof Array) {
			for (var i=0; i<x.length; i++) {
				if (this.lookahead.type === x[i]) {
					return true;
				}
			}
			return false;
		}

		return this.lookahead.type === x;
	},

	error: function (expecting) {
		throw new Error('Očekáváno "' + expecting + '", nalezeno "' + this.lookahead + '"');
	},

	rule: function (name) {
		var r = new RuleNode(name);
		if (this.root === null) {
			this.root = r;
		} else {
			this.currentNode.addChild(r, this.currentNode);
		}
		var tmp = this.currentNode;
		this.currentNode = r;

		// spusteni pravidla
		this[name]();

		this.currentNode = tmp;
	},

	parse: function (input) {
		this.lexer.input(input);
		this.consume();
		this.rule('style'); // startovaci symbol
		this.dfs(this.root);
		return this.styles;
	},

	dfs: function (root) {
		this.inspectNode(root);
		if (root.children) {
			for (var i=0; i<root.children.length; i++) {
				this.dfs(root.children[i]);
			}
		}
	},

	inspectNode: function (node) {
		if (node instanceof TokenNode) {
			var tokenType = node.token.type;
			var tokenValue = node.token.value;

			// def - zacatek
			if (tokenType === Lexer.KW_DEF) {
				this.currentStyle = {};
				this.currentStyle.font = {};
				this.currentStyle.par = {};
				this.defParamFound = false;
			}

			else if (tokenType === Lexer.IDENTIFIER) {
				// nazev stylu
				if (node.parent.name === 'def') {
					this.currentStyle.name = node.token.value.trim().substring(1); // remove '\'
				}

				// velikost pisma
				else if (tokenValue.contains(['tiny', 'scriptsize', 'footnotesize', 'small', 'large', 'Large', 'LARGE', 'huge', 'Huge'])) {
					this.currentStyle.font.size = tokenValue.substring(1);
				}
			}

			else if (tokenType === Lexer.TEXT) {
				// zarovnani
				if (node.parent.parent.name === 'environment') {
					this.currentStyle.par.align = node.token.value;
				}

				// font face
				else if (node.parent.parent.children[0].token.value.contains('fontspec')) {
					this.currentStyle.font.face = node.token.value;
				}
			}

			// rez pisma
			else if (tokenType === Lexer.KW_SHAPE) {
				if (tokenValue.contains('textsc')) {
					this.currentStyle.font.caps = true;
				}
				else if (tokenValue.contains('texttt')) {
					this.currentStyle.font.type = true;
				}
				else if (tokenValue.contains(['textit', 'textbf', 'textsl'])) {
					this.currentStyle.font.style = tokenValue.trim().substring(1); // remove '\'
				}
			}

			// def parametr
			else if (tokenType === Lexer.DEF_PARAM) {
				if (node.parent.name === 'defContent') {
					this.defParamFound = true;
				}
			}

			// def - ukonceni
			else if (tokenType === Lexer.RIGHT_CBRACKET) {
				if (node.parent.name === 'def') {
					if (this.currentStyle) {
						var style = $.extend(true, {}, this.defaultStyle, this.currentStyle);
						this.styles.push(style); // ulozeni stylu
					}
				}
			}
		}
		else if (node instanceof RuleNode) {
			if (node.name === 'dimension') {
				// skip
				if (node.parent.name === 'skipParameter') {
					var skipType = node.parent.parent.children[0].token.value;
					var skipDimension = '' + node.children[0].token.value + node.children[1].token.value;
					if (skipType.contains('left')) {
						this.currentStyle.par.hspaceLeft = skipDimension;
					} else if (skipType.contains('right')) {
						this.currentStyle.par.hspaceRight = skipDimension;
					}
				}

				// space
				if (node.parent.name === 'spaceParameter') {
					var spaceType = (this.defParamFound) ? 'after' : 'before';
					var spaceDimension = '' + node.children[0].token.value + node.children[1].token.value;
					if (spaceType === 'before') {
						this.currentStyle.par.vspaceBefore = spaceDimension;
					} else {
						this.currentStyle.par.vspaceAfter = spaceDimension;
					}
				}
			}
		}
	},

	style: function () {
		if (this.isLookahead(Lexer.KW_DEF)) {
			this.rule('def');
			this.rule('style');
		}
	},

	def: function () {
		if (this.isLookahead(Lexer.KW_DEF)) {
			this.expect(Lexer.KW_DEF);
			this.expect(Lexer.IDENTIFIER);
			this.expect(Lexer.DEF_PARAM);
			this.expect(Lexer.LEFT_CBRACKET);
			this.rule('defContent');
			this.expect(Lexer.RIGHT_CBRACKET);
		} else {
			this.error('\\def');
		}
	},

	defContent: function () {
		if (this.isLookahead(Lexer.DEF_PARAM)) {
			this.expect(Lexer.DEF_PARAM);
			this.rule('defContent');
		}
		else if (this.isLookahead([Lexer.IDENTIFIER, Lexer.KW_BEGIN, Lexer.KW_BEGINGROUP, Lexer.KW_FONTSPEC, Lexer.KW_SHAPE, Lexer.KW_SKIP, Lexer.KW_SPACE, Lexer.LEFT_CBRACKET])) {
			this.rule('command');
			this.rule('defContent');
		}
		else if (this.isLookahead([Lexer.KW_END, Lexer.KW_ENDGROUP, Lexer.RIGHT_CBRACKET])) {
			return;
		}
		else {
			this.error('def - obsah');
		}
	},

	command: function () {
		if (this.isLookahead(Lexer.IDENTIFIER)) {
			this.expect(Lexer.IDENTIFIER);
		}
		else if (this.isLookahead(Lexer.KW_BEGIN)) {
			this.rule('environment');
		}
		else if (this.isLookahead(Lexer.KW_BEGINGROUP)) {
			this.rule('group');
		}
		else if (this.isLookahead(Lexer.KW_FONTSPEC)) {
			this.expect(Lexer.KW_FONTSPEC);
			this.rule('textParameter');
		}
		else if (this.isLookahead(Lexer.KW_SHAPE)) {
			this.expect(Lexer.KW_SHAPE);
			this.rule('shapeParameter');
		}
		else if (this.isLookahead(Lexer.KW_SKIP)) {
			this.expect(Lexer.KW_SKIP);
			this.rule('skipParameter');
		}
		else if (this.isLookahead(Lexer.KW_SPACE)) {
			this.expect(Lexer.KW_SPACE);
			this.rule('spaceParameter');
		}
		else if (this.isLookahead(Lexer.LEFT_CBRACKET)) {
			this.rule('group');
		}
		else {
			this.error('příkaz');
		}
	},

	environment: function () {
		if (this.isLookahead(Lexer.KW_BEGIN)) {
			this.expect(Lexer.KW_BEGIN);
			this.rule('textParameter');
			this.rule('defContent');
			this.expect(Lexer.KW_END);
			this.rule('textParameter');
		} else {
			this.error('\\begin');
		}
	},

	textParameter: function () {
		if (this.isLookahead(Lexer.LEFT_CBRACKET)) {
			this.expect(Lexer.LEFT_CBRACKET);
			this.expect(Lexer.TEXT);
			this.expect(Lexer.RIGHT_CBRACKET);
		} else {
			this.error('{');
		}
	},

	group: function () {
		if (this.isLookahead(Lexer.KW_BEGINGROUP)) {
			this.expect(Lexer.KW_BEGINGROUP);
			this.rule('defContent');
			this.expect(Lexer.KW_ENDGROUP);
		}
		else if (this.isLookahead(Lexer.LEFT_CBRACKET)) {
			this.expect(Lexer.LEFT_CBRACKET);
			this.rule('defContent');
			this.expect(Lexer.RIGHT_CBRACKET);
		}
		else {
			this.error('\\begingroup nebo {');
		}
	},

	shapeParameter: function () {
		if (this.isLookahead(Lexer.LEFT_CBRACKET)) {
			this.expect(Lexer.LEFT_CBRACKET);
			this.rule('bracketsContent');
			this.expect(Lexer.RIGHT_CBRACKET);
		} else {
			this.error('{');
		}
	},

	bracketsContent: function () {
		if (this.isLookahead(Lexer.DEF_PARAM)) {
			this.expect(Lexer.DEF_PARAM);
			this.rule('bracketsContent');
		}
		else if (this.isLookahead([Lexer.IDENTIFIER, Lexer.KW_BEGIN, Lexer.KW_BEGINGROUP, Lexer.KW_SHAPE, Lexer.KW_SKIP, Lexer.KW_SPACE, Lexer.LEFT_CBRACKET, Lexer.KW_FONTSPEC])) {
			this.rule('command');
			this.rule('bracketsContent');
		}
		else if (this.isLookahead(Lexer.RIGHT_CBRACKET)) {
			return;
		}
		else {
			this.error('obsah bloku');
		}
	},

	skipParameter: function () {
		if (this.isLookahead(Lexer.EQUALS)) {
			this.expect(Lexer.EQUALS);
			this.rule('dimension');
		} else {
			this.error('=');
		}
	},

	dimension: function () {
		if (this.isLookahead(Lexer.NUMBER)) {
			this.expect(Lexer.NUMBER);
			this.expect(Lexer.UNIT);
		} else {
			this.error('číslo');
		}
	},

	spaceParameter: function () {
		if (this.isLookahead(Lexer.LEFT_CBRACKET)) {
			this.expect(Lexer.LEFT_CBRACKET);
			this.rule('dimension');
			this.expect(Lexer.RIGHT_CBRACKET);
		} else {
			this.error('{');
		}
	}
};


/**
 * ParseTree class
 * @constructor
 */
var ParseTree = function () {
	this.parent = null;
	this.children = null;
};

ParseTree.prototype = {
	addChild: function (arg, parent) {
		if ($.type(arg) === 'string') {
			var ruleNode = new RuleNode(arg);
			this.addChild(ruleNode, parent);
			return ruleNode;
		}

		if (arg instanceof Token) {
			var tokenNode = new TokenNode(arg);
			this.addChild(tokenNode, parent);
			return tokenNode;
		}

		if (arg instanceof ParseTree) {
			if (this.children === null) {
				this.children = [];
			}
			arg.parent = parent;
			this.children.push(arg);
		}
	}
};


/**
 * RuleNode class
 * @constructor
 */
var RuleNode = function (name) {
	this.name = name;
};

RuleNode.prototype = new ParseTree();

RuleNode.prototype.toString = function () {
	return 'Rule: ' + this.name;
};


/**
 * TokenNode class
 * @constructor
 */
var TokenNode = function (token) {
	this.token = token;
};

TokenNode.prototype = new ParseTree();

TokenNode.prototype.toString = function () {
	return 'Token: ' + this.token.value;
};


// Util functions
String.prototype.contains = function (arg) {
	if (arg instanceof Array) {
		for (var i=0; i<arg.length; i++) {
			if (this.indexOf(arg[i]) > -1) {
				return true;
			}
		}
		return false;
	} else {
		return this.indexOf(arg) > -1;
	}
};
